#include <fcntl.h>
#include <sched.h>
#include <semaphore.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/wait.h>


void processus_fils(int prio)
{
	sem_t * sem;
	sem = sem_open("/solution-07-04", O_RDWR);
	sem_wait(sem);
	fprintf(stderr, "[%d] (priorite %d) obtient le semaphore\n",
	                 getpid(), prio);
	usleep(10000);
	fprintf(stderr, "[%d] (priorite %d) lache le semaphore\n",
	                 getpid(), prio);
	sem_post(sem);
	exit(EXIT_SUCCESS);
}



int main(int argc, char * argv [])
{
	int i;
	pid_t pid;
	sem_t * sem;
	struct sched_param param;
	
	// Creer d'un nouveau semaphore a 1 unite.
	sem_unlink("/solution-07-04");
	sem = sem_open("/solution-07-04", O_RDWR | O_CREAT, 0600, 1);
	// Verrouiller le semaphore
	sem_wait(sem);
	
	fprintf(stderr, "Lancement de 5 processus de priorites croissantes\n");
	for (i = 0; i < 5; i ++) {
		param.sched_priority = (i+1) * 10;
		pid = fork();
		if (pid == 0) {
			processus_fils((i+1) * 10);
			exit(EXIT_SUCCESS);
		}
		if (sched_setscheduler(pid, SCHED_FIFO, & param) != 0) {
			perror("sched_setscheduler");
			kill(pid, SIGKILL);
			exit(EXIT_FAILURE);
		}
	}
	sleep(1);
	fprintf(stderr, "Liberation du semaphore\n");
	sem_post(sem);
	
	// Attente de la fin des cinq processus
	for (i = 0; i < 5; i ++)
		waitpid(-1, NULL, 0);
	// Reprise du semaphore
	sem_wait(sem);
	
	fprintf(stderr, "Lancement de 5 processus de priorites decroissantes\n");
	for (i = 0; i < 5; i ++) {
		param.sched_priority = 60 - 10*(i+1);
		pid = fork();
		if (pid == 0) {
			processus_fils(60 - 10*(i+1));
			exit(EXIT_SUCCESS);
		}
		if (sched_setscheduler(pid, SCHED_FIFO, & param) != 0) {
			perror("sched_setscheduler");
			kill(pid, SIGKILL);
			exit(EXIT_FAILURE);
		}
	}
	sleep(1);
	fprintf(stderr, "Liberation du semaphore\n");
	sem_post(sem);
	
	// Attente de la fin des cinq processus
	for (i = 0; i < 5; i ++)
		waitpid(-1, NULL, 0);
	// Reprise du semaphore
	sem_wait(sem);
	
	fprintf(stderr, "Lancement de 5 processus RR de meme priorite\n");
	for (i = 0; i < 5; i ++) {
		param.sched_priority = 30;
		pid = fork();
		if (pid == 0) {
			processus_fils(30);
			exit(EXIT_SUCCESS);
		}
		if (sched_setscheduler(pid, SCHED_RR, & param) != 0) {
			perror("sched_setscheduler");
			kill(pid, SIGKILL);
			exit(EXIT_FAILURE);
		}
	}
	sleep(1);
	fprintf(stderr, "Liberation du semaphore\n");
	sem_post(sem);
	
	// Attente de la fin des cinq processus
	for (i = 0; i < 5; i ++)
		waitpid(-1, NULL, 0);
	
	return EXIT_SUCCESS;
}


