#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/mman.h>

#include <rtdk.h>
#include <native/task.h>
#include <native/timer.h>


void fonction_periodique (void * arg)
{
	RTIME precedent = 0; 
	RTIME heure = 0;
	RTIME periode;
	RTIME duree;
	long long int ecart;
	long long int ecart_max = 0;
	
	periode = (RTIME) arg; // en centaines de usec
	periode = periode * 100000; // en ns
	rt_task_set_periodic(NULL, TM_NOW, periode);
	while(1) {
		rt_task_wait_period(NULL);
		heure = rt_timer_read();
		if (precedent != 0) { 
			duree = heure - precedent;
			ecart = duree - periode;
			ecart = ecart / 1000; 
			if (ecart > ecart_max)
				ecart_max = ecart;
			rt_fprintf(stderr, "[%d] ecart=%3lld  ecart_max=%lld\n", (int) arg, ecart, ecart_max);
		}
		precedent = heure;
	}
}


int main(void)
{
	int i;
	int nb_cpus;
	int err;
	RT_TASK * tasks;
	char taskname[32];
	
	nb_cpus = sysconf(_SC_NPROCESSORS_ONLN);
	mlockall(MCL_CURRENT|MCL_FUTURE);
	rt_print_auto_init(1);
	
	tasks = calloc(nb_cpus, sizeof(RT_TASK));
	if (tasks == NULL) {
		perror("calloc");
		return EXIT_FAILURE;
	}
	
	for (i = 0; i < nb_cpus; i ++) {
		sprintf(taskname, "solution-10-04-%d\n", i);
		if ((err = rt_task_spawn(& (tasks[i]), taskname, 0, 99, T_JOINABLE | T_CPU(i),
		                         fonction_periodique, (void *) (i+1))) != 0) {
			fprintf(stderr, "rt_task_spawn: %s\n", strerror(-err));
			exit(EXIT_FAILURE);
		}
	}
	for (i = 0; i < nb_cpus; i ++) {
		rt_task_join(& (tasks[i]));
	}
	return EXIT_SUCCESS;
}

