#include <linux/version.h>
#include <linux/device.h>
#include <linux/module.h>
#include <linux/mutex.h>
#include <linux/sched.h>
#include <linux/cdev.h>
#include <linux/fs.h>
#include <asm/uaccess.h>

static dev_t          exemple_dev;
static struct cdev    exemple_cdev;
static struct class * exemple_class = NULL;

static int exemple_read (struct file * filp, char * buffer,
                          size_t length, loff_t * offset);

static int exemple_write(struct file * filp, const char * buffer,
                        size_t lg, loff_t * offset);

static struct file_operations fops_exemple = {
	.owner   =  THIS_MODULE,
	.read    =  exemple_read,
	.write   =  exemple_write,
};

#define NB_MINEURS 1

static int __init exemple_init (void)
{
	int erreur;
	
	erreur = alloc_chrdev_region(& exemple_dev,
	                             0, NB_MINEURS, THIS_MODULE->name);
	if (erreur < 0)
		return erreur;
		
	exemple_class = class_create(THIS_MODULE, "solution");
	if (IS_ERR(exemple_class)) {
		unregister_chrdev_region(exemple_dev, NB_MINEURS);
		return -EINVAL;
	}
	device_create(exemple_class, NULL, exemple_dev,
	              NULL, THIS_MODULE->name);

	cdev_init(& exemple_cdev, & fops_exemple);

	erreur = cdev_add(& exemple_cdev, exemple_dev, NB_MINEURS);
	if (erreur != 0) {
		device_destroy(exemple_class, exemple_dev);
		class_destroy(exemple_class);
		unregister_chrdev_region(exemple_dev, NB_MINEURS);
		return erreur;
	}
	return 0; 
}



static void __exit exemple_exit (void)
{
	cdev_del(& exemple_cdev);
	device_destroy(exemple_class, exemple_dev);
	class_destroy(exemple_class);
	unregister_chrdev_region(exemple_dev, NB_MINEURS);
}



typedef struct element_file {
	char * contenu;
	size_t lg;
	struct element_file * suivant;
	struct element_file * precedent;
};

struct element_file * premier = NULL;

DEFINE_MUTEX(mtx_file);


static int exemple_read(struct file * filp, char * buffer,
                        size_t lg, loff_t * offset)
{
	struct element_file * elm;
	
	if (premier == NULL)
		return 0;
	if (* offset > 0)
		return 0;
		
	if (mutex_lock_interruptible(& mtx_file) != 0)
		return -ERESTARTSYS;
	
	if (lg > premier->lg)
		lg = premier->lg;
	
	copy_to_user(buffer, premier->contenu, lg);
	
	elm = premier->suivant;
	vfree(premier->contenu);
	vfree (premier);
	premier = elm;
	mutex_unlock(& mtx_file);
	*offset = lg;
	return lg;
}


static int exemple_write(struct file * filp, const char * buffer,
                        size_t lg, loff_t * offset)
{
	struct element_file * elm, *new;
	
	new = vmalloc(sizeof(struct element_file));
	if (new == NULL)
		return -ENOMEM;
	new->contenu = vmalloc(lg);
	if (new->contenu == NULL)
		return -ENOMEM;
	copy_from_user(new->contenu, buffer, lg);
	new->lg = lg;
	new->suivant = NULL;
	
	if (mutex_lock_interruptible(& mtx_file) != 0)
		return -ERESTARTSYS;

	if (premier == NULL) {
			premier = new;
			new->precedent = NULL;
	} else {
		for (elm = premier; elm->suivant != NULL; elm = elm->suivant)
			;
		elm->suivant = new;
		new->precedent = elm;
	}
	
	mutex_unlock(& mtx_file);
	return lg;
}

module_init(exemple_init);
module_exit(exemple_exit);
MODULE_LICENSE("GPL");

